# Constants unique to the fs backend.
# the flag to pass for a dry run to make sure syntax is proper

DRY_RUN_FLAG = "--clippy-validate"
# the flag to pass to get detailed help for constructing the class
HELP_FLAG = "--clippy-help"

# Arguments for execution progress
PROGRESS_START_KEY = "progress_start"
PROGRESS_END_KEY = "progress_end"
PROGRESS_INC_KEY = "progress_inc"
PROGRESS_SET_KEY = "progress_set"
